#include <DebugOverSoftSerial.h>
#include <SimpleSim800.h>
#include <LowPower.h>

#define LED_RED_PIN          8
#define LED_GREEN_PIN        9
#define LED_BLUE_PIN        13
#define LED_DELAY_BLINK_MS  65

#define SWITCH_ONE_PIN       2
#define SWITCH_TWO_PIN       3

#define BURN_READING_COUNT   8

// ~5 minutes / 8 seconds
// duration of loop ist a little bit longer than 8 seconds
// so we chose a 10% lower value
#define INITIAL_TIME_CHECK_CYCLE_FACTOR   35
// ~6 hours / 8 seconds
// duration of loop ist a little bit longer than 8 seconds
// so we chose a 10% lower value
#define LONG_TIME_CHECK_CYCLE_FACTOR    2430
// ~1 hour / 8 seconds
#define TIME_CHECK_CYCLE_FACTOR_HOUR     450

#define SLEEP_PERIOD  SLEEP_8S

int BATTERY_SENSE_PIN = A3;  // select the input pin for the battery sense point
/*
 * Calcolation for Voltage divider
 * low resistor = 2.2kOhm
 * high resistor = 10kOhm
 * reference voltage = 1.1V
 * claculation
 * Vpin = Vmeasure * 2.2 / (2.2 + 10)
 * Vpin = 1.1V * Xadc / 1023
 * Vmeasure = (1.1V / 1023) * ((2.2 + 10) / 2.2) * Xadc
 * Vmeasure = 0.005963V * Xadc
 */
const float VOLTAGE_CALCULAION_FACTOR = 5.963;

const byte messageTimes[] = {0,6,12,18};

DebugOverSoftSerial logger;
// APN and SIM are configured directly in library
SimpleSim800 sim800;

const char url[] = "http://yourdomain/traps/ping";
byte nextMessageTimeIndex = 0;
int timeCheckCycleFactor = 0;
byte actualHour = 0;
char imei[20] = {0};

byte lastSwitchOneValue = -1;
byte lastSwitchTwoValue = -1;

int loopCounter = 0;

void setup() 
{
  // set some ports
  pinMode(SWITCH_ONE_PIN, INPUT);
  pinMode(SWITCH_TWO_PIN, INPUT);

  pinMode(LED_RED_PIN, OUTPUT);
  pinMode(LED_GREEN_PIN, OUTPUT);
  pinMode(LED_BLUE_PIN, OUTPUT);
  
  // set reference for voltage measurement
  analogReference(INTERNAL);

  // start GSM module
  sim800.begin();
  blinkGreen(1);
  // this delay is necessary, it helps the device to be ready and connect to a network
  delay(5000); 

  initializeGprs();   
  sim800.getImei(imei);

  actualHour = hourWithTimezone(false);
  for (byte i = 0; i < sizeof(messageTimes); i++)
  {
    int messageTimeAtIndex = messageTimes[i];
    if (messageTimeAtIndex > actualHour)
    {
      nextMessageTimeIndex = i;
      break;
    }
  }

  // lookup of time will happen about every 5 minutes
  timeCheckCycleFactor = INITIAL_TIME_CHECK_CYCLE_FACTOR;
}

void loop() 
{
  // put your main code here, to run repeatedly:

  // kind of heartbeat
  blinkGreen(1);

  // read the volage
  analogReference(INTERNAL);
  burnReadings(BATTERY_SENSE_PIN, BURN_READING_COUNT);
  delay(10); 
  int result = analogRead(BATTERY_SENSE_PIN);
  float voltage = VOLTAGE_CALCULAION_FACTOR * result;

  // read the switches
  byte actualSwitchOneValue = digitalRead(SWITCH_ONE_PIN);
  byte actualSwitchTwoValue = 0;

  // do we need to report becaus of switch changes?
  if(actualSwitchOneValue != lastSwitchOneValue ||
     actualSwitchTwoValue != lastSwitchTwoValue)
  {
    if (reportToServer((int)voltage, actualSwitchOneValue, actualSwitchTwoValue) == 0)
    {
      lastSwitchOneValue = actualSwitchOneValue;
      lastSwitchTwoValue = actualSwitchTwoValue;
    }    
  }

  // we don't need to check the time every cycle
  // assume a time of 8s per cycle
  if (loopCounter >= timeCheckCycleFactor)
  {
    // lookup of time will happen about every 5 minutes
    timeCheckCycleFactor = INITIAL_TIME_CHECK_CYCLE_FACTOR;
    
    actualHour = hourWithTimezone(true);
    
    if (actualHour >= 0 && actualHour < 24) 
    {
      int messageTimeAtIndex = messageTimes[nextMessageTimeIndex];

      /*
      logger.print(F("actual hour is "));
      logger.print(String(actualHour));
      logger.print(F(" messageTime at index is "));
      logger.println(String(messageTimeAtIndex));
      */
      
      if (actualHour == messageTimeAtIndex)
      {
        // first we increment the index to the next message-time
        nextMessageTimeIndex++;
        if (nextMessageTimeIndex >= sizeof(messageTimes))
        {
          nextMessageTimeIndex = 0;
        }  
        // then we report
        reportToServer((int)voltage, actualSwitchOneValue, actualSwitchTwoValue); 

        // we can set the timeCheckCycleFactor to a value
        // that the lookup will happen just about before the next hour
        timeCheckCycleFactor = LONG_TIME_CHECK_CYCLE_FACTOR;
      }
      else if ((messageTimeAtIndex > 0 && actualHour + 1 < messageTimeAtIndex) ||
               (messageTimeAtIndex == 0 && actualHour == 22))
      {
        timeCheckCycleFactor = TIME_CHECK_CYCLE_FACTOR_HOUR;
      }
    }
    loopCounter = 0;

    // logger.print(F("timeCheckCycleFactor is "));
    // logger.println(String(timeCheckCycleFactor));
  }

  loopCounter++;
  LowPower.powerDown(SLEEP_PERIOD, ADC_OFF, BOD_ON);
}

void initializeGprs()
{
  if(sim800.init() != 0) {
      delay(1000);
      blinkRed(2);
  }

  if(sim800.connect() != 0) {
      delay(1000);
      blinkRed(3);
  }
  delay(3000); 
  blinkGreen(2);
}

int reportToServer(int voltage, byte switchOne, byte switchTwo)
{
  digitalWrite(LED_BLUE_PIN, HIGH);
  char voltageString[5];
  itoa(voltage, voltageString, 10);
  
  char postData[50] = {0};
  // append IMEI
  strcat(postData, "imei=");
  strcat(postData, imei);

  // append voltage
  strcat(postData, "&mv=");
  strcat(postData, voltageString);

      // append switch-values
  if (switchOne == 0)
  {
    strcat(postData, "&s1=0");
  }
  else
  {
    strcat(postData, "&s1=1");
  }
    
  if (switchTwo == 0)
  {
    strcat(postData, "&s2=0");
  }
  else
  {
    strcat(postData, "&s2=1");
  }

  // wake up GSM-module
  if(sim800.connect() == 0)
  {
    sim800.postData(url, postData);
    sim800.disconnect();
    digitalWrite(LED_BLUE_PIN, LOW);
    return 0;
  }
  digitalWrite(LED_BLUE_PIN, LOW);
  blinkRed(10);
  return 1;
}

void blinkGreen(byte amount)
{
  for (byte i = 0; i < amount; i++)
  {
    blink(LED_GREEN_PIN, LED_DELAY_BLINK_MS);
  }
}

void blinkRed(byte amount)
{
  for (byte i = 0; i < amount; i++)
  {
    blink(LED_RED_PIN, LED_DELAY_BLINK_MS);
  }
}

void blink(byte pin, byte delayTime)
{
  pinMode(pin, OUTPUT);
  digitalWrite(pin,HIGH);
  delay(delayTime);
  digitalWrite(pin,LOW);
  delay(delayTime);
}

byte hourWithTimezone(bool withPowerSaving)
{
  if (withPowerSaving)
  {
    sim800.disablePowerSaving();
  }
  
  // Answer of SIM800L looks like
  // yy/MM/dd,hh:mm:ss+/-zz
  int year, month, day, hour, minute, second;
  char dateString[22] = {0};

  if (sim800.getDateTime(dateString) == GET_TIME_SUCCESS)
  {
    if (withPowerSaving)
    {
      sim800.enablePowerSaving();
    }
    
    sscanf(dateString, "%d/%d/%d,%d:%d:%d+00", &year, &month, &day, &hour, &minute, &second);
    year = year + 2000;
    // Starting with MET
    byte offset = 1;
    if(sim800.isSummertime(year, month, day, hour, offset))
    {
      // Seems we've got MEST
      offset = 2;
    }
    if (hour + offset > 23)
    {
      return hour + offset - 24;
    }
    return hour + offset;
  }
  if (withPowerSaving)
  {
    sim800.enablePowerSaving();
  }
  
  logger.println(F("Main: Getting time failed"));
  return 255; // the time will be sent later in response to serial mesg
}

void burnReadings(int pin, int counter)
{
  for (int i = 0; i < counter; i++)
  {
    analogRead(pin);
  }
}

/**
 * returns volatge in mV
 */
static int readVoltage()
{
  int result;
  
  analogReference(INTERNAL);
  burnReadings(BATTERY_SENSE_PIN, BURN_READING_COUNT);
  delay(10); 
  result = analogRead(BATTERY_SENSE_PIN);
  float voltage = VOLTAGE_CALCULAION_FACTOR * result;
  return (int)voltage;
}
